/*
 * abiathar - rtu.c
 *
 * Modbus RTU framing support
 *
 * Copyright (C) 2019  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <kernel.h>
#include <tty.h>
#include <crc16.h>
#include <timer.h>
#include <rtu.h>

#define RTU_CHECKSUM_INIT	0xffff

#define RTU_MAX 1

#define RTU_MTU 256

#define RTU_T15 ((75 * HZ + 99999) / 100000)
#define RTU_T35 ((175 * HZ + 99999) / 100000)

#define RTU_BUF_SIZE	8

enum rtu_state {
	RTU_INIT = 0,
	RTU_IDLE,
	RTU_TX,
	RTU_RX,
	RTU_WAIT,
};

struct rtu {
	struct tty_struct *tty;
	unsigned int len;
	void (*callback)(struct rtu *p, const unsigned char *frame,
			unsigned int len);
	u8 frame[RTU_MTU];
	u16 crc;
	unsigned long rx_jiffy;
	unsigned int addr;
	enum rtu_state state; 
};

static struct rtu rtu_table[RTU_MAX];

#define le32_to_cpu(x)	(x)
#define cpu_to_le32(x)	(x)

static void rtu_receive_buf(struct tty_struct *tty, const unsigned char *cp,
		char *fp, int count)
	__attribute__((hot));

static void rtu_receive_buf(struct tty_struct *tty, const unsigned char *cp,
		char *fp, int count)
{
	struct rtu *p = tty->disc_data;
	unsigned int len = p->len;
	u16 crc = p->crc;

	if (!count)
		return;

	if (p->state == RTU_WAIT)
		goto drop;

	if (p->state != RTU_RX && p->state != RTU_IDLE)
		return;

	if (p->state == RTU_IDLE)
		p->state = RTU_RX;

	p->rx_jiffy = jiffies;
	do {
		unsigned char c;
		u8 *frame = p->frame;

		/*
		 * if case of receive error drop character and current frame
		 */
		if (fp && *fp++)
			goto drop;

		/*
		 * drop current frame if frame is too large
		 */
		if (len >= RTU_MTU)
			goto drop;

		c = *cp;
		frame[len++] = c;
		crc = crc16_byte(crc, c);
		cp++;
		count--;
	} while (count > 0);
	p->len = len;
	p->crc = crc;
	return;
drop:
	p->len = 0;
	p->crc = RTU_CHECKSUM_INIT;
}

void rtu_poll(struct rtu *p)
{
	unsigned long t = jiffies;

	switch (p->state) {
	case RTU_INIT:
	case RTU_TX:
		if (time_after(t, p->rx_jiffy + RTU_T35))
			p->state = RTU_IDLE;
		break;
	case RTU_WAIT:
		if (time_after(t, p->rx_jiffy + RTU_T35)) {
			if (p->len > 2 && !p->crc)
				p->callback(p, p->frame, p->len - 2);
			p->len = 0;
			p->crc = RTU_CHECKSUM_INIT;
			p->state = RTU_IDLE;
		}
		break;
	case RTU_RX:
		if (time_after(t, p->rx_jiffy + RTU_T15))
			p->state = RTU_WAIT;
		break;
	case RTU_IDLE:
		break;
	}
}

int rtu_xmit(struct rtu *p, unsigned char *frame, unsigned int len)
{
	unsigned int crc;

	crc = crc16(RTU_CHECKSUM_INIT, frame, len);
	frame[len + 0] = (crc >> 0) & 0xff;
	frame[len + 1] = (crc >> 8) & 0xff;
	p->tty->ops->write(p->tty, frame, len + 2);
	return 0;
}

static const struct tty_ldisc_ops rtu_ldisc_ops = {
	.receive_buf		= rtu_receive_buf,
};

struct rtu *rtu_init(struct tty_struct *tty,
		void (*callback)(struct rtu *p, const unsigned char *frame,
			unsigned int len))
{
	struct rtu *p;

	for (p = rtu_table; p < rtu_table + RTU_MAX; p++) {
		if (!p->tty) {
			p->callback = callback;
			p->tty = tty;
			p->state = RTU_INIT;
			p->rx_jiffy = jiffies;
			p->len = 0;
			p->crc = RTU_CHECKSUM_INIT;

			tty_set_ldisc(tty, &rtu_ldisc_ops, p);
			return p;
		}
	}
	return NULL;
}
