/*
 * grader-io - uart-nxp.c
 *
 * NXP LPC111x serial driver
 *
 * Copyright (C) 2016  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <kernel.h>
#include <asm_io.h>
#include <serial_reg.h>
#include <uart-nxp.h>
#include <tty.h>

#define UART_LPC_ACR		8	/* 0x20 */
#define UART_FDR		10	/* 0x28 */
#define UART_TER		12	/* 0x30 */

#define NVIC_BASE	0xe000e100
#define NVIC_ISER	0x0

#define TX_FIFO_SIZE		16
static void uart_transmit_chars(struct uart *p) __attribute__((hot));

static void uart_transmit_chars(struct uart *p)
{
	unsigned int i = TX_FIFO_SIZE;
	unsigned int tail = p->tx_tail;
	unsigned int head = p->tx_head;
	unsigned int x;

	if (tail == head) {
		p->ier &= ~UART_IER_THRI;
		writel(p->ier, p->regs + UART_IER);
		return;
	}

	while ((tail & 3) || ((tail >> 2) == (head >> 2))) {
		const u8 *b8 = (const u8 *) p->tx_buf;

		x = b8[tail];
		tail = (tail + 1) & (UART_TX_BUF_SIZE - 1);
		writel(x, p->regs + UART_TX);
		if (tail == head)
			goto out;
		i--;
	}

	if ((tail >> 2) != (head >> 2)) {
		u32 __iomem *regs = p->regs;
		const u32 *buf = p->tx_buf;

		do {
			x = buf[tail >> 2];
			writel((x >> 0) & 0xff, regs + UART_TX);
			writel((x >> 8) & 0xff, regs + UART_TX);
			writel((x >> 16) & 0xff, regs + UART_TX);
			writel((x >> 24) & 0xff, regs + UART_TX);
			tail = (tail + 4) & (UART_TX_BUF_SIZE - 1);
			if ((tail >> 2) == (head >> 2))
				goto out;
			i -= 4;
		} while (i >= 4);
	}

out:
	p->tx_tail = tail;
}

static int uart_tty_write(struct tty_struct *tty, const unsigned char *buf,
		int count)
{
	struct uart *p = container_of(tty, struct uart, tty);
	u8 *tx_buf = (u8 *) p->tx_buf;
	int ret = 0;
	unsigned int tail;
	unsigned int head;
	unsigned int next;
	unsigned long flags;

	local_irq_save(flags);
	tail = p->tx_tail;
	head = p->tx_head;
	while (count > 0) {
		next = (head + 1) & (UART_TX_BUF_SIZE - 1);
		if (next == tail)
			break;
		tx_buf[head] = *buf++;
		head = next;
		count--;
		ret++;
	}
	p->tx_head = head;

	if ((p->tx_head != p->tx_tail) && !(p->ier & UART_IER_THRI)) {
		p->ier |= UART_IER_THRI;
		writel(p->ier, p->regs + UART_IER);

		/*
		 * workaround for buggy TX IIR status
		 */
		uart_transmit_chars(p);
	}
	local_irq_restore(flags);
	return ret;
}

static int uart_tty_write_room(struct tty_struct *tty)
{
	struct uart *p = container_of(tty, struct uart, tty);
	unsigned long flags;
	unsigned int tail;
	unsigned int head;

	local_irq_save(flags);
	tail = p->tx_tail;
	head = p->tx_head;
	local_irq_restore(flags);

	if (tail > head)
		return tail - head - 1;
	else
		return UART_TX_BUF_SIZE + tail - head - 1;
}

static inline unsigned int uart_decode_flag(unsigned int lsr)
{
	unsigned int flag = 0;

	if (lsr & UART_LSR_BI)
		flag = TTY_BREAK;
	else if (lsr & UART_LSR_PE)
		flag = TTY_PARITY;
	else if (lsr & UART_LSR_FE)
		flag = TTY_FRAME;
	return flag;
}

void uart_irq_handler(struct uart *p)
{
	u32 __iomem *regs = p->regs;
	unsigned int isr;
	unsigned int lsr;
	unsigned int data;
	u32 cp[UART_RX_BUF_SIZE / 4];
	u32 fp[UART_RX_BUF_SIZE / 4];

	do {
		lsr = readl(regs + UART_LSR);
		if (lsr & UART_LSR_DR) {
			unsigned int count = 0;
			unsigned int flags = 0;

			do {
				data = readl(regs + UART_RX);
				flags = 0;
				if (unlikely(lsr & UART_LSR_BRK_ERROR_BITS))
					flags = uart_decode_flag(lsr);
				lsr = readl(regs + UART_LSR);
				if (!(lsr & UART_LSR_DR))
					goto loop_out;

				count++;
				data |= readl(regs + UART_RX) << 8;
				if (unlikely(lsr & UART_LSR_BRK_ERROR_BITS))
					flags |= uart_decode_flag(lsr) << 8;
				lsr = readl(regs + UART_LSR);
				if (!(lsr & UART_LSR_DR))
					goto loop_out;

				count++;
				data |= readl(regs + UART_RX) << 16;
				if (unlikely(lsr & UART_LSR_BRK_ERROR_BITS))
					flags |= uart_decode_flag(lsr) << 16;
				lsr = readl(regs + UART_LSR);
				if (!(lsr & UART_LSR_DR))
					goto loop_out;

				count++;
				data |= readl(regs + UART_RX) << 24;
				if (unlikely(lsr & UART_LSR_BRK_ERROR_BITS))
					flags |= uart_decode_flag(lsr) << 24;
				lsr = readl(regs + UART_LSR);

loop_out:
				cp[count >> 2] = data;
				fp[count >> 2] = flags;

				count++;
				if (!(lsr & UART_LSR_DR))
					break;
			} while (count < 4);
			tty_receive_buf(&p->tty, (unsigned char *) cp,
					(char *) fp, count);
		}

		if (unlikely((lsr & UART_LSR_THRE) && (p->ier & UART_IER_THRI)))
			uart_transmit_chars(p);

		isr = readl(regs + UART_IIR);
	} while ((isr & UART_IIR_NO_INT) == 0);

}

static const struct tty_operations uart_tty_operations = {
	.write		= uart_tty_write,
	.write_room	= uart_tty_write_room,
};

void uart_init(struct uart *p, unsigned long addr, unsigned int irq,
		unsigned int div)
{
	u32 lcr;
	u32 x;

	p->tty.ops = &uart_tty_operations;
	p->regs = ioremap(addr, 1024);

	lcr = UART_LCR_WLEN8 | UART_LCR_DLAB | UART_LCR_PARITY | UART_LCR_EPAR;
	writel(lcr, p->regs + UART_LCR);

	writel(0x10, p->regs + UART_FDR);
	writel(0, p->regs + UART_DLM);
	writel(div, p->regs + UART_DLL);

	lcr &= ~UART_LCR_DLAB;
	writel(lcr, p->regs + UART_LCR);

	/*
	 * set FIFO trigger level to 8 bytes, 37500 interrupts per second
	 * at 3Mbps
	 */
	writel(UART_FCR_R_TRIG_10 | UART_FCR_ENABLE_FIFO, p->regs + UART_FCR);

	p->ier = UART_IER_RDI;
	writel(p->ier, p->regs + UART_IER);

	x = readl((u32 *)(NVIC_BASE + NVIC_ISER));
	x |= 1UL << irq;
	writel(x, (u32 *)(NVIC_BASE + NVIC_ISER));
}
